from __future__ import annotations

__all__ = ["StripStringSeriesTransformer"]


from pandas import Series

from flamme.transformer.series.base import BaseSeriesTransformer


class StripStringSeriesTransformer(BaseSeriesTransformer):
    r"""Implements a transformer to strip the strings in a
    ``pandas.Series``.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.series import StripString
        >>> transformer = StripString()
        >>> transformer
        StripStringSeriesTransformer()
        >>> series = pd.Series(["a ", " b", "  c  ", " d ", "e"])
        >>> series = transformer.transform(series)
        >>> series
        0    a
        1    b
        2    c
        3    d
        4    e
        dtype: object
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def transform(self, series: Series) -> Series:
        return series.apply(lambda x: x.strip() if isinstance(x, str) else x)



from __future__ import annotations

__all__ = ["EmptySection"]

from collections.abc import Sequence

from flamme.section.base import BaseSection


class EmptySection(BaseSection):
    r"""Implements an empty section.

    This section is implemented to deal with missing columns or to skip
    some analyses.
    """

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        return ""

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return ""



from __future__ import annotations

__all__ = ["SectionDict"]

from collections.abc import Sequence

from coola.utils import str_indent

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)


class SectionDict(BaseSection):
    r"""Implements a section to manage a dictionary of sections.

    Args:
        sections (dict): Specifies the dictionary of sections.
    """

    def __init__(self, sections: dict[str, BaseSection]) -> None:
        self._sections = sections

    @property
    def sections(self) -> dict[str, BaseSection]:
        return self._sections

    def get_statistics(self) -> dict:
        return {name: section.get_statistics() for name, section in self._sections.items()}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        report = []
        if tags:
            report.append(
                f'<h{valid_h_tag(depth+1)} id="{tags2id(tags)}">{number} '
                f"{tags2title(tags)} </h{valid_h_tag(depth+1)}>"
            )
            report.append(GO_TO_TOP)
            report.append('<p style="margin-top: 1rem;">')
        for i, (name, section) in enumerate(self._sections.items()):
            report.append(
                section.render_html_body(
                    number=f"{number}{i + 1}.", tags=list(tags) + [name], depth=depth + 1
                )
            )
        return "\n".join(report)

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        toc = []
        if tags:
            toc.append(render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth))
        subtoc = []
        for i, (name, section) in enumerate(self._sections.items()):
            line = section.render_html_toc(
                number=f"{section}{i+1}.",
                tags=list(tags) + [name],
                depth=depth + 1,
                max_depth=max_depth,
            )
            if line:
                subtoc.append(f"  {str_indent(line)}")
        if subtoc:
            toc.extend(["<ul>", "\n".join(subtoc), "</ul>"])
        return "\n".join(toc)



from __future__ import annotations

__all__ = [
    "find_columns_decimal",
    "find_columns_str",
    "find_columns_type",
]

from decimal import Decimal

from pandas import DataFrame

from flamme.utils.dtype import df_column_types


def find_columns_type(df: DataFrame, cls: type) -> tuple[str, ...]:
    r"""Finds the list of columns that contains a given type.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame.
        cls: Specifies the type to find.

    Returns:
        tuple: The tuple of columns with the given type.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.utils.filtering import find_columns_type
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["1", "2", "3", "4", "5"],
        ...         "col4": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> find_columns_type(df, str)
        ('col2', 'col3', 'col4')
    """
    types = df_column_types(df)
    return tuple(col for col, tps in types.items() if cls in tps)


def find_columns_decimal(df: DataFrame) -> tuple[str, ...]:
    r"""Finds the list of columns that contains the type string.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame.

    Returns:
        tuple: The tuple of columns with the type string.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from decimal import Decimal
        >>> from flamme.utils.filtering import find_columns_decimal
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, Decimal(5)],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["1", "2", "3", "4", "5"],
        ...         "col4": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> find_columns_decimal(df)
        ('col1',)
    """
    return find_columns_type(df, Decimal)


def find_columns_str(df: DataFrame) -> tuple[str, ...]:
    r"""Finds the list of columns that contains the type string.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame.

    Returns:
        tuple: The tuple of columns with the type string.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.utils.filtering import find_columns_str
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["1", "2", "3", "4", "5"],
        ...         "col4": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> find_columns_str(df)
        ('col2', 'col3', 'col4')
    """
    return find_columns_type(df, str)



from __future__ import annotations

__all__ = [
    "BaseSection",
    "ColumnContinuousAdvancedSection",
    "ColumnContinuousSection",
    "ColumnDiscreteSection",
    "ColumnTemporalContinuousSection",
    "ColumnTemporalDiscreteSection",
    "ColumnTemporalNullValueSection",
    "DataTypeSection",
    "DuplicatedRowSection",
    "EmptySection",
    "GlobalTemporalNullValueSection",
    "MarkdownSection",
    "MostFrequentValuesSection",
    "NullValueSection",
    "SectionDict",
    "TemporalNullValueSection",
]

from flamme.section.base import BaseSection
from flamme.section.continuous import ColumnContinuousSection
from flamme.section.continuous_advanced import ColumnContinuousAdvancedSection
from flamme.section.continuous_temporal import ColumnTemporalContinuousSection
from flamme.section.discrete import ColumnDiscreteSection
from flamme.section.discrete_temporal import ColumnTemporalDiscreteSection
from flamme.section.dtype import DataTypeSection
from flamme.section.duplicate import DuplicatedRowSection
from flamme.section.empty import EmptySection
from flamme.section.mapping import SectionDict
from flamme.section.markdown import MarkdownSection
from flamme.section.most_frequent import MostFrequentValuesSection
from flamme.section.null import NullValueSection, TemporalNullValueSection
from flamme.section.null_temp_col import ColumnTemporalNullValueSection
from flamme.section.null_temp_global import GlobalTemporalNullValueSection



from __future__ import annotations

__all__ = ["DecimalToNumericDataFrameTransformer"]


import pandas as pd
from pandas import DataFrame
from tqdm import tqdm

from flamme.transformer.df.base import BaseDataFrameTransformer
from flamme.utils.filtering import find_columns_decimal


class DecimalToNumericDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implements a ``pandas.DataFrame`` to convert all the columns with
    ``Decimal`` objects to floats.

    Args:
        **kwargs: Specifies the keyword arguments for
            ``pandas.to_numeric``.

    Example usage:

    .. code-block:: pycon

        >>> from decimal import Decimal
        >>> import pandas as pd
        >>> from flamme.transformer.df import DecimalToNumeric
        >>> transformer = DecimalToNumericDataFrameTransformer()
        >>> transformer
        DecimalToNumericDataFrameTransformer()
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, Decimal(5)],
        ...         "col2": [Decimal(1), Decimal(2), Decimal(3), Decimal(4), Decimal(5)],
        ...         "col3": ["1", "2", "3", "4", "5"],
        ...         "col4": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> df.dtypes
        col1    object
        col2    object
        col3    object
        col4    object
        dtype: object
        >>> df = transformer.transform(df)
        >>> df.dtypes
        col1     int64
        col2   float64
        col3    object
        col4    object
        dtype: object
    """

    def __init__(self, **kwargs) -> None:
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = ", ".join([f"{key}={value}" for key, value in self._kwargs.items()])
        return f"{self.__class__.__qualname__}({args})"

    def transform(self, df: DataFrame) -> DataFrame:
        columns = find_columns_decimal(df)
        for col in tqdm(columns, desc="Converting Decimal to numeric type"):
            df[col] = pd.to_numeric(df[col], **self._kwargs)
        return df



from __future__ import annotations

__all__ = ["ColumnDataFrameTransformer"]

import logging
from collections.abc import Mapping

from coola.utils import str_indent, str_mapping
from pandas import DataFrame

from flamme.transformer.df.base import BaseDataFrameTransformer
from flamme.transformer.series.base import BaseSeriesTransformer

logger = logging.getLogger(__name__)


class ColumnDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implements a ``pandas.DataFrame`` transformer that applies
    ``pandas.Series`` transformers on some columns.

    Args:
        columns (``Mapping``): Specifies the ``pandas.Series``
            transformers.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.df import Column
        >>> from flamme.transformer.series import ToNumeric, StripString
        >>> transformer = Column({"col2": ToNumeric(), "col3": StripString()})
        >>> transformer
        ColumnDataFrameTransformer(
          (col2): ToNumericSeriesTransformer()
          (col3): StripStringSeriesTransformer()
          (ignore_missing): False
        )
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": [" a", "b ", " c ", "  d  ", "e"],
        ...     }
        ... )
        >>> df = transformer.transform(df)
        >>> df
           col1  col2 col3
        0     1     1    a
        1     2     2    b
        2     3     3    c
        3     4     4    d
        4     5     5    e
    """

    def __init__(
        self, columns: Mapping[str, BaseSeriesTransformer | dict], ignore_missing: bool = False
    ) -> None:
        self._columns = columns
        self._ignore_missing = bool(ignore_missing)

    def __repr__(self) -> str:
        columns = self._columns | {"ignore_missing": self._ignore_missing}
        args = f"\n  {str_indent(str_mapping(columns))}\n"
        return f"{self.__class__.__qualname__}({args})"

    def transform(self, df: DataFrame) -> DataFrame:
        for col, transformer in self._columns.items():
            if col not in df:
                if self._ignore_missing:
                    logger.warning(
                        f"Skipping transformation for column {col} because the column is missing"
                    )
                else:
                    raise RuntimeError(
                        f"Column {col} is not in the DataFrame (columns:{sorted(df.columns)})"
                    )
            else:
                df[col] = transformer.transform(df[col])
        return df



from __future__ import annotations

__all__ = ["MarkdownAnalyzer"]

from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import MarkdownSection


class MarkdownAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer that adds a mardown string to the report..

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import MarkdownAnalyzer
        >>> analyzer = MarkdownAnalyzer(desc="hello cats!")
        >>> analyzer
        MarkdownAnalyzer()
        >>> df = pd.DataFrame({})
        >>> section = analyzer.analyze(df)
    """

    def __init__(self, desc: str) -> None:
        self._desc = str(desc)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def analyze(self, df: DataFrame) -> MarkdownSection:
        return MarkdownSection(desc=self._desc)



from __future__ import annotations

__all__ = ["BaseSeriesTransformer", "is_series_transformer_config", "setup_series_transformer"]

import logging
from abc import ABC

from objectory import AbstractFactory
from objectory.utils import is_object_config
from pandas import Series

logger = logging.getLogger(__name__)


class BaseSeriesTransformer(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to transform a ``pandas.Series``.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.series import ToNumeric
        >>> transformer = ToNumeric()
        >>> transformer
        ToNumericSeriesTransformer()
        >>> series = pd.Series(["1", "2", "3", "4", "5"])
        >>> series.dtype
        dtype('O')
        >>> series = transformer.transform(series)
        >>> series.dtype
        dtype('int64')
    """

    def transform(self, df: Series) -> Series:
        r"""Transforms a ``pandas.Series``.

        Args:
        ----
            df (``pandas.Series``): Specifies the ``pandas.Series``
                to transform.

        Returns:
        -------
            ``pandas.Series``: The transformed ``pandas.Series``.

        Example usage:

        .. code-block:: pycon

            >>> import pandas as pd
            >>> from flamme.transformer.series import ToNumeric
            >>> transformer = ToNumeric()
            >>> series = pd.Series(["1", "2", "3", "4", "5"])
            >>> series = transformer.transform(series)
            >>> series.dtype
            dtype('int64')
        """


def is_series_transformer_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseSeriesTransformer``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config (dict): Specifies the configuration to check.

    Returns:
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseSeriesTransformer`` object.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.transformer.series import is_series_transformer_config
        >>> is_series_transformer_config({"_target_": "flamme.transformer.series.ToNumeric"})
        True
    """
    return is_object_config(config, BaseSeriesTransformer)


def setup_series_transformer(
    transformer: BaseSeriesTransformer | dict,
) -> BaseSeriesTransformer:
    r"""Sets up a ``pandas.Series`` transformer.

    The transformer is instantiated from its configuration
    by using the ``BaseSeriesTransformer`` factory function.

    Args:
        transformer (``BaseSeriesTransformer`` or dict): Specifies a
            ``pandas.Series`` transformer or its configuration.

    Returns:
        ``BaseSeriesTransformer``: An instantiated transformer.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.transformer.series import setup_series_transformer
        >>> transformer = setup_series_transformer(
        ...     {"_target_": "flamme.transformer.series.ToNumeric"}
        ... )
        >>> transformer
        ToNumericSeriesTransformer()
    """
    if isinstance(transformer, dict):
        logger.info("Initializing a series transformer from its configuration... ")
        transformer = BaseSeriesTransformer.factory(**transformer)
    if not isinstance(transformer, BaseSeriesTransformer):
        logger.warning(
            f"transformer is not a `BaseSeriesTransformer` (received: {type(transformer)})"
        )
    return transformer



from __future__ import annotations

__all__ = ["human_file_size", "sanitize_path"]

from pathlib import Path
from urllib.parse import unquote, urlparse

from flamme.utils.format import human_byte


def human_file_size(path: Path | str, decimal: int = 2) -> str:
    r"""Gets a human-readable representation of a file size.

    Args:
        path (``pathlib.Path`` or str): Specifies the file.
        decimal (int, optional): Specifies the number of decimal
            digits. Default: ``2``

    Returns:
        str: The file size in a human-readable format.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.utils.path import human_file_size
        >>> human_file_size("README.md")
        '...B'
    """
    return human_byte(size=sanitize_path(path).stat().st_size, decimal=decimal)


def sanitize_path(path: Path | str) -> Path:
    r"""Sanitizes a given path.

    Args:
        path (``pathlib.Path`` or str): Specifies the path to
            sanitize.

    Returns:
        ``pathlib.Path``: The sanitized path.

    Example usage:

    .. code-block:: pycon

        >>> from pathlib import Path
        >>> from flamme.utils.path import sanitize_path
        >>> sanitize_path("something")
        PosixPath('.../something')
        >>> sanitize_path("")
        PosixPath('...')
        >>> sanitize_path(Path("something"))
        PosixPath('.../something')
        >>> sanitize_path(Path("something/./../"))
        PosixPath('...')
    """
    if isinstance(path, str):
        # Use urlparse to parse file URI: https://stackoverflow.com/a/15048213
        path = Path(unquote(urlparse(path).path))
    return path.expanduser().resolve()



from __future__ import annotations

__all__ = ["GlobalTemporalNullValueSection"]

import logging
from collections.abc import Sequence

import numpy as np
from jinja2 import Template
from matplotlib import pyplot as plt
from pandas import DataFrame

from flamme.section.base import BaseSection
from flamme.section.null import plot_temporal_null_total
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels

logger = logging.getLogger(__name__)


class GlobalTemporalNullValueSection(BaseSection):
    r"""Implements a section to analyze the temporal distribution of null
    values for all columns.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        figsize (``tuple`` or list , optional): Specifies the figure
            size in inches. The first dimension is the width and the
            second is the height. Default: ``None``
    """

    def __init__(
        self,
        df: DataFrame,
        dt_column: str,
        period: str,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        if dt_column not in df:
            raise ValueError(
                f"Datetime column {dt_column} is not in the DataFrame "
                f"(columns:{sorted(df.columns)})"
            )

        self._df = df
        self._dt_column = dt_column
        self._period = period
        self._figsize = figsize

    @property
    def df(self) -> DataFrame:
        r"""``pandas.DataFrame``: The DataFrame to analyze."""
        return self._df

    @property
    def dt_column(self) -> str:
        r"""str: The datetime column."""
        return self._dt_column

    @property
    def period(self) -> str:
        r"""str: The temporal period used to analyze the data."""
        return self._period

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""tuple: The individual figure size in pixels. The first
        dimension is the width and the second is the height."""
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            "Rendering the temporal distribution of null values for all columns "
            f"| datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "dt_column": self._dt_column,
                "figure": create_temporal_null_figure(
                    df=self._df,
                    dt_column=self._dt_column,
                    period=self._period,
                    figsize=self._figsize,
                ),
                "table": create_temporal_null_table(
                    df=self._df,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of null values in all columns.
The column <em>{{dt_column}}</em> is used as the temporal column.

{{figure}}

{{table}}
<p style="margin-top: 1rem;">
"""


def create_temporal_null_figure(
    df: DataFrame,
    dt_column: str,
    period: str,
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Creates a HTML representation of a figure with the temporal null
    value distribution.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        figsize (``tuple`` or list , optional): Specifies the figure
            size in inches. The first dimension is the width and the
            second is the height. Default: ``None``

    Returns:
        str: The HTML representation of the figure.
    """
    if df.shape[0] == 0:
        return ""

    num_nulls, total, labels = prepare_data(df=df, dt_column=dt_column, period=period)

    fig, ax = plt.subplots(figsize=figsize)
    plot_temporal_null_total(ax=ax, labels=labels, num_nulls=num_nulls, total=total)
    readable_xticklabels(ax, max_num_xticks=100)
    return figure2html(fig, close_fig=True)


def create_temporal_null_table(df: DataFrame, dt_column: str, period: str) -> str:
    r"""Creates a HTML representation of a table with the temporal
    distribution of null values.

    Args:
        df (``DataFrame``): Specifies the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.

    Returns:
        str: The HTML representation of the table.
    """
    if df.shape[0] == 0:
        return ""
    num_nulls, totals, labels = prepare_data(df=df, dt_column=dt_column, period=period)
    rows = []
    for label, num_null, total in zip(labels, num_nulls, totals):
        rows.append(create_temporal_null_table_row(label=label, num_nulls=num_null, total=total))
    return Template(
        """
<details>
    <summary>Statistics per period</summary>

    <p>The following table shows some statistics for each period.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr>
                <th>period</th>
                <th>number of null values</th>
                <th>number of non-null values</th>
                <th>total number of values</th>
                <th>percentage of null values</th>
                <th>percentage of non-null values</th>
            </tr>
        </thead>
        <tbody class="tbody table-group-divider">
            {{rows}}
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render({"rows": "\n".join(rows), "period": period})


def create_temporal_null_table_row(label: str, num_nulls: int, total: int) -> str:
    r"""Creates the HTML code of a new table row.

    Args:
        row ("pd.core.frame.Pandas"): Specifies a DataFrame row.

    Returns:
        str: The HTML code of a row.
    """
    num_non_nulls = total - num_nulls
    return Template(
        """<tr>
    <th>{{label}}</th>
    <td {{num_style}}>{{num_nulls}}</td>
    <td {{num_style}}>{{num_non_nulls}}</td>
    <td {{num_style}}>{{total}}</td>
    <td {{num_style}}>{{num_nulls_pct}}</td>
    <td {{num_style}}>{{num_non_nulls_pct}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "label": label,
            "num_nulls": f"{num_nulls:,}",
            "num_non_nulls": f"{num_non_nulls:,}",
            "total": f"{total:,}",
            "num_nulls_pct": f"{100 * num_nulls / total:.2f}%",
            "num_non_nulls_pct": f"{100 * num_non_nulls / total:.2f}%",
        }
    )


def prepare_data(
    df: DataFrame,
    dt_column: str,
    period: str,
) -> tuple[np.ndarray, np.ndarray, list]:
    r"""Prepares the data to create the figure and table.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.

    Returns:
        tuple: A tuple with 3 values. The first value is a numpy NDArray
            that contains the number of null values per period. The
            second value is a numpy NDArray that contains the total
            number of values. The third value is a list that contains
            the label of each period.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.section.null_temp_global import prepare_data
        >>> num_nulls, total, labels = prepare_data(
        ...     df=pd.DataFrame(
        ...         {
        ...             "col1": np.array([np.nan, 1, 0, 1]),
        ...             "col2": np.array([np.nan, 1, 0, np.nan]),
        ...             "datetime": pd.to_datetime(
        ...                 ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
        ...             ),
        ...         }
        ...     ),
        ...     dt_column="datetime",
        ...     period="M",
        ... )
        >>> num_nulls
        array([2, 0, 0, 1])
        >>> total
        array([2, 2, 2, 2])
        >>> labels
        ['2020-01', '2020-02', '2020-03', '2020-04']
    """
    df = df.copy()
    columns = df.columns.tolist()
    columns.remove(dt_column)
    dt_col = "__datetime__"
    df[dt_col] = df[dt_column].dt.to_period(period)
    df.loc[:, columns] = df[columns].isnull()

    df_num_nulls = df.groupby(dt_col)[columns].sum().sum(axis=1).sort_index()
    df_total = df.groupby(dt_col)[columns].count().sum(axis=1).sort_index()
    labels = [str(dt) for dt in df_num_nulls.index]
    return df_num_nulls.to_numpy().astype(int), df_total.to_numpy().astype(int), labels



from __future__ import annotations

__all__ = ["ColumnTemporalDiscreteSection"]

import logging
from collections.abc import Sequence

import numpy as np
from jinja2 import Template
from matplotlib import pyplot as plt
from pandas import DataFrame

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels
from flamme.utils.sorting import mixed_typed_sort

logger = logging.getLogger(__name__)


class ColumnTemporalDiscreteSection(BaseSection):
    r"""Implements a section that analyzes the temporal distribution of a
    column with discrete values.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column of the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        figsize (``tuple`` or ``None``, optional): Specifies the figure
            size in inches. The first dimension is the width and the
            second is the height. Default: ``None``
    """

    def __init__(
        self,
        df: DataFrame,
        column: str,
        dt_column: str,
        period: str,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._df = df
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._figsize = figsize

    @property
    def column(self) -> str:
        return self._column

    @property
    def dt_column(self) -> str:
        return self._dt_column

    @property
    def period(self) -> str:
        return self._period

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""tuple: The individual figure size in pixels. The first
        dimension is the width and the second is the height."""
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            f"Analyzing the temporal discrete distribution of {self._column} | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "dt_column": self._dt_column,
                "period": self._period,
                "figure": create_temporal_figure(
                    df=self._df,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                    figsize=self._figsize,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of column <em>{{column}}</em>
by using the column <em>{{dt_column}}</em>.

{{figure}}

{{table}}
<p style="margin-top: 1rem;">
"""


def create_temporal_figure(
    df: DataFrame,
    column: str,
    dt_column: str,
    period: str,
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Creates a HTML representation of a figure with the temporal value
    distribution.

    Args:
        df (``DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        log_y (bool, optional): If ``True``, it represents the bars
            with a log scale. Default: ``False``
        figsize (``tuple`` or ``None``, optional): Specifies the figure
            size in inches. The first dimension is the width and the
            second is the height. Default: ``None``

    Returns:
        str: The HTML representation of the figure.
    """
    if df.shape[0] == 0:
        return "<span>&#9888;</span> No figure is generated because the column is empty"
    df = df[[column, dt_column]].copy()
    col_dt, col_count = "__datetime__", "__count__"
    df[col_dt] = df[dt_column].dt.to_period(period).astype(str)
    df = df[[column, col_dt]].groupby(by=[col_dt, column], dropna=False)[column].size()
    df = DataFrame({col_count: df}).reset_index().sort_values(by=[col_dt, column])
    df = df.set_index([col_dt, column]).unstack(fill_value=0)

    labels = mixed_typed_sort(df[col_count].columns.tolist())
    num_labels = len(labels)
    steps = df.index.tolist()
    x = np.arange(len(steps), dtype=int)
    bottom = np.zeros_like(x)
    width = 0.9 if len(steps) < 50 else 1
    fig, ax = plt.subplots(figsize=figsize)
    my_cmap = plt.get_cmap("viridis")
    for i, label in enumerate(labels):
        count = df[col_count][label].to_numpy()
        ax.bar(x, count, label=label, bottom=bottom, width=width, color=my_cmap(i / num_labels))
        bottom += count

    if len(labels) <= 10:
        ax.legend()
    ax.set_xticks(x, labels=steps)
    readable_xticklabels(ax, max_num_xticks=100)
    ax.set_xlim(-0.5, len(steps) - 0.5)
    ax.set_ylabel("Number of occurrences")
    ax.set_title(f"Temporal distribution for column {column}")
    return figure2html(fig, close_fig=True)



from __future__ import annotations

__all__ = ["NullValueSection", "TemporalNullValueSection"]

import logging
import math
from collections.abc import Sequence

import numpy as np
from jinja2 import Template
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from pandas import DataFrame

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels

logger = logging.getLogger(__name__)


class NullValueSection(BaseSection):
    r"""Implements a section that analyzes the number of null values.

    Args:
        columns (``Sequence``): Specifies the column names.
        null_count (``numpy.ndarray``): Specifies the number of null
            values for each column.
        total_count (``numpy.ndarray``): Specifies the total number
            of values for each column.
        figsize (``tuple`` or ``None``, optional): Specifies the figure
            size in inches. The first dimension is the width and the
            second is the height. Default: ``None``
    """

    def __init__(
        self,
        columns: Sequence[str],
        null_count: np.ndarray,
        total_count: np.ndarray,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._columns = tuple(columns)
        self._null_count = null_count.flatten().astype(int)
        self._total_count = total_count.flatten().astype(int)
        self._figsize = figsize

        if len(self._columns) != self._null_count.shape[0]:
            raise RuntimeError(
                f"columns ({len(self._columns):,}) and null_count ({self._null_count.shape[0]:,}) "
                "do not match"
            )
        if len(self._columns) != self._total_count.shape[0]:
            raise RuntimeError(
                f"columns ({len(self._columns):,}) and total_count "
                f"({self._total_count.shape[0]:,}) do not match"
            )

    @property
    def columns(self) -> tuple[str, ...]:
        r"""Tuple: The columns used to compute the duplicated rows."""
        return self._columns

    @property
    def null_count(self) -> np.ndarray:
        r"""``numpy.ndarray``: The number of null values for each
        column."""
        return self._null_count

    @property
    def total_count(self) -> np.ndarray:
        r"""``numpy.ndarray``: The total number of values for each
        column."""
        return self._total_count

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""tuple: The individual figure size in pixels. The first
        dimension is the width and the second is the height."""
        return self._figsize

    def get_statistics(self) -> dict:
        return {
            "columns": self._columns,
            "null_count": tuple(self._null_count.tolist()),
            "total_count": tuple(self._total_count.tolist()),
        }

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info("Rendering the null value distribution of all columns...")
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "table_alpha": self._create_table(sort_by="column"),
                "table_sort": self._create_table(sort_by="null"),
                "bar_figure": self._create_bar_figure(),
                "num_columns": f"{len(self._columns):,}",
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the number and proportion of null values for the {{num_columns}}
columns.
In the following histogram, the columns are sorted by ascending order of null values.


{{bar_figure}}

<details>
    <summary>Show analysis per column</summary>

    <p style="margin-top: 1rem;">
    The following tables show the number and proportion of null values for the {{num_columns}}
    columns.
    The background color of the row indicates the proportion of missing values:
    dark blues indicates more missing values than light blues.

    <div class="container-fluid">
        <div class="row align-items-start">
            <div class="col align-self-center">
                <p><b>Columns sorted by alphabetical order</b></p>

                {{table_alpha}}

            </div>
            <div class="col">
                <p><b>Columns sorted by ascending order of missing values</b></p>

                {{table_sort}}

            </div>
        </div>
    </div>
</details>
<p style="margin-top: 1rem;">
"""

    def _create_bar_figure(self) -> str:
        df = self._get_dataframe().sort_values(by="null")
        fig, ax = plt.subplots(figsize=self._figsize)
        labels = df["column"].tolist()
        ax.bar(x=labels, height=df["null"].to_numpy(), color="tab:blue")
        if labels:
            ax.set_xlim(-0.5, len(labels) - 0.5)
        readable_xticklabels(ax, max_num_xticks=100)
        ax.set_xlabel("column")
        ax.set_ylabel("number of null values")
        ax.set_title("number of null values per column")
        return figure2html(fig, close_fig=True)

    def _create_table(self, sort_by: str) -> str:
        df = self._get_dataframe().sort_values(by=sort_by)
        rows = "\n".join(
            [
                create_table_row(column=column, null_count=null_count, total_count=total_count)
                for column, null_count, total_count in zip(
                    df["column"].to_numpy(), df["null"].to_numpy(), df["total"].to_numpy()
                )
            ]
        )
        return Template(
            """
<table class="table table-hover table-responsive w-auto" >
    <thead class="thead table-group-divider">
        <tr>
            <th>column</th>
            <th>null pct</th>
            <th>null count</th>
            <th>total count</th>
        </tr>
    </thead>
    <tbody class="tbody table-group-divider">
        {{rows}}
        <tr class="table-group-divider"></tr>
    </tbody>
</table>
"""
        ).render({"rows": rows})

    def _get_dataframe(self) -> DataFrame:
        return DataFrame(
            {"column": self._columns, "null": self._null_count, "total": self._total_count}
        )


def create_table_row(column: str, null_count: int, total_count: int) -> str:
    r"""Creates the HTML code of a new table row.

    Args:
        column (str): Specifies the column name.
        null_count (int): Specifies the number of null values.
        total_count (int): Specifies the total number of rows.

    Returns:
        str: The HTML code of a row.
    """
    pct = null_count / total_count
    return Template(
        """<tr>
    <th style="background-color: rgba(0, 191, 255, {{null_pct}})">{{column}}</th>
    <td {{num_style}}>{{null_pct}}</td>
    <td {{num_style}}>{{null_count}}</td>
    <td {{num_style}}>{{total_count}}</td>
</tr>"""
    ).render(
        {
            "num_style": f'style="text-align: right; background-color: rgba(0, 191, 255, {pct})"',
            "column": column,
            "null_count": f"{null_count:,}",
            "null_pct": f"{pct:.4f}",
            "total_count": f"{total_count:,}",
        }
    )


class TemporalNullValueSection(BaseSection):
    r"""Implements a section to analyze the temporal distribution of null
    values.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        ncols (int, optional): Specifies the number of columns.
            Default: ``2``
        figsize (``tuple``, optional): Specifies the figure size in
            inches. The first dimension is the width and the second is
            the height. Default: ``(7, 5)``
    """

    def __init__(
        self,
        df: DataFrame,
        dt_column: str,
        period: str,
        ncols: int = 2,
        figsize: tuple[float, float] = (7, 5),
    ) -> None:
        self._df = df
        self._dt_column = dt_column
        self._period = period
        self._ncols = ncols
        self._figsize = figsize

    @property
    def df(self) -> DataFrame:
        r"""``pandas.DataFrame``: The DataFrame to analyze."""
        return self._df

    @property
    def dt_column(self) -> str:
        r"""str: The datetime column."""
        return self._dt_column

    @property
    def period(self) -> str:
        r"""str: The temporal period used to analyze the data."""
        return self._period

    @property
    def ncols(self) -> int:
        r"""int: The number of columns to show the figures."""
        return self._ncols

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""Tuple or ``None``: The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            "Rendering the temporal null value distribution of all columns | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._dt_column,
                "figure": create_temporal_null_figure(
                    df=self._df,
                    dt_column=self._dt_column,
                    period=self._period,
                    ncols=self._ncols,
                    figsize=self._figsize,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the monthly distribution of null values.
The column <em>{{column}}</em> is used as the temporal column.

{{figure}}
"""


def create_temporal_null_figure(
    df: DataFrame,
    dt_column: str,
    period: str,
    ncols: int = 2,
    figsize: tuple[float, float] = (7, 5),
) -> str:
    r"""Creates a HTML representation of a figure with the temporal null
    value distribution.

    Args:
        df (``DataFrame``): Specifies the DataFrame to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        ncols (int, optional): Specifies the number of columns.
            Default: ``2``
        figsize (``tuple``, optional): Specifies the figure size in
            inches. The first dimension is the width and the second is
            the height. Default: ``(7, 5)``

    Returns:
        str: The HTML representation of the figure.
    """
    if df.shape[0] == 0:
        return ""
    columns = sorted([col for col in df.columns if col != dt_column])
    nrows = math.ceil(len(columns) / ncols)
    fig, axes = plt.subplots(
        nrows=nrows, ncols=ncols, figsize=(figsize[0] * ncols, figsize[1] * nrows)
    )

    for i, column in enumerate(columns):
        if ncols > 1:
            ax = axes[i // ncols, i % ncols]
        else:
            ax = axes[i]
        ax.set_title(f"column: {column}")

        num_nulls, total, labels = prepare_data(
            df=df, column=column, dt_column=dt_column, period=period
        )
        plot_temporal_null_total(ax=ax, labels=labels, num_nulls=num_nulls, total=total)
        readable_xticklabels(ax, max_num_xticks=100 // ncols)

    return figure2html(fig, close_fig=True)


def plot_temporal_null_total(
    ax: Axes, num_nulls: np.ndarray, total: np.ndarray, labels: list
) -> None:
    color = "tab:blue"
    x = np.arange(len(labels))
    ax.set_ylabel("number of null/total values", color=color)
    ax.tick_params(axis="y", labelcolor=color)
    ax.bar(x=x, height=total, color="tab:cyan", alpha=0.5, label="total")
    ax.bar(x=x, height=num_nulls, color=color, alpha=0.8, label="null")
    ax.legend()

    ax2 = ax.twinx()
    color = "black"
    ax2.set_ylabel("percentage", color=color)
    ax2.tick_params(axis="y", labelcolor=color)
    ax2.plot(x, num_nulls / total, "o-", color=color)

    ax.set_xticks(x, labels=labels)
    ax.set_xlim(-0.5, len(labels) - 0.5)


def prepare_data(
    df: DataFrame,
    column: str,
    dt_column: str,
    period: str,
) -> tuple[np.ndarray, np.ndarray, list]:
    r"""Prepares the data to create the figure and table.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.

    Returns:
        tuple: A tuple with 3 values. The first value is a numpy NDArray
            that contains the number of null values per period. The
            second value is a numpy NDArray that contains the total
            number of values. The third value is a list that contains
            the label of each period.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.section.null_temp_col import prepare_data
        >>> num_nulls, total, labels = prepare_data(
        ...     df=pd.DataFrame(
        ...         {
        ...             "col": np.array([np.nan, 1, 0, 1]),
        ...             "datetime": pd.to_datetime(
        ...                 ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
        ...             ),
        ...         }
        ...     ),
        ...     column="col",
        ...     dt_column="datetime",
        ...     period="M",
        ... )
        >>> num_nulls
        array([1, 0, 0, 0])
        >>> total
        array([1, 1, 1, 1])
        >>> labels
        ['2020-01', '2020-02', '2020-03', '2020-04']
    """
    df = df[[column, dt_column]].copy()
    dt_col = "__datetime__"
    df[dt_col] = df[dt_column].dt.to_period(period)

    null_col = f"__{column}_isnull__"
    df.loc[:, null_col] = df.loc[:, column].isnull()

    df_num_nulls = df.groupby(dt_col)[null_col].sum().sort_index()
    df_total = df.groupby(dt_col)[null_col].count().sort_index()
    labels = [str(dt) for dt in df_num_nulls.index]
    return df_num_nulls.to_numpy().astype(int), df_total.to_numpy().astype(int), labels



from __future__ import annotations

__all__ = ["ColumnContinuousSection"]

import logging
from collections.abc import Sequence

import numpy as np
from jinja2 import Template
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from pandas import Series

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    auto_yscale_continuous,
    compute_statistics,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.array import nonnan
from flamme.utils.figure import figure2html, readable_xticklabels
from flamme.utils.range import find_range

logger = logging.getLogger(__name__)


class ColumnContinuousSection(BaseSection):
    r"""Implements a section that analyzes a continuous distribution of
    values.

    Args:
        series: Specifies the series/column to analyze.
        column: Specifies the column name.
        nbins: Specifies the number of bins in the histogram.
        yscale: Specifies the y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        xmin: Specifies the minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: Specifies the maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: Specifies the figure size in inches. The first
            dimension is the width and the second is the height.
    """

    def __init__(
        self,
        series: Series,
        column: str,
        nbins: int | None = None,
        yscale: str = "auto",
        xmin: float | str | None = None,
        xmax: float | str | None = None,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._series = series
        self._column = column
        self._nbins = nbins
        self._yscale = yscale
        self._xmin = xmin
        self._xmax = xmax
        self._figsize = figsize

    @property
    def column(self) -> str:
        return self._column

    @property
    def yscale(self) -> str:
        return self._yscale

    @property
    def nbins(self) -> int | None:
        return self._nbins

    @property
    def series(self) -> Series:
        return self._series

    @property
    def xmin(self) -> float | str | None:
        return self._xmin

    @property
    def xmax(self) -> float | str | None:
        return self._xmax

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""tuple: The individual figure size in pixels. The first
        dimension is the width and the second is the height."""
        return self._figsize

    def get_statistics(self) -> dict[str, float | int]:
        return compute_statistics(self._series)

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(f"Rendering the continuous distribution of {self._column}")
        stats = self.get_statistics()
        null_values_pct = (
            f"{100 * stats['num_nulls'] / stats['count']:.2f}" if stats["count"] > 0 else "N/A"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "table": create_stats_table(stats=stats, column=self._column),
                "total_values": f"{stats['count']:,}",
                "unique_values": f"{stats['nunique']:,}",
                "null_values": f"{stats['num_nulls']:,}",
                "null_values_pct": null_values_pct,
                "histogram_figure": create_histogram_figure(
                    series=self._series,
                    column=self._column,
                    stats=stats,
                    nbins=self._nbins,
                    yscale=self._yscale,
                    xmin=self._xmin,
                    xmax=self._xmax,
                    figsize=self._figsize,
                ),
                "boxplot_figure": create_boxplot_figure(
                    series=self._series,
                    xmin=self._xmin,
                    xmax=self._xmax,
                    figsize=self._figsize,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the discrete distribution of values for column <em>{{column}}</em>.

<ul>
  <li> total values: {{total_values}} </li>
  <li> number of unique values: {{unique_values}} </li>
  <li> number of null values: {{null_values}} / {{total_values}} ({{null_values_pct}}%) </li>
</ul>

{{histogram_figure}}
{{boxplot_figure}}
{{table}}
<p style="margin-top: 1rem;">
"""


def create_boxplot_figure(
    series: Series,
    xmin: float | str | None = None,
    xmax: float | str | None = None,
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Creates the HTML code of a boxplot figure.

    Args:
        series: Specifies the series/column to analyze.
        xmin: Specifies the minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: Specifies the maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: Specifies the figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        str: The HTML code of the figure.
    """
    array = series.dropna().to_numpy()
    if array.size == 0:
        return "<span>&#9888;</span> No figure is generated because the column is empty"
    xmin, xmax = find_range(array, xmin=xmin, xmax=xmax)
    if figsize is not None:
        figsize = (figsize[0], figsize[0] / 10)
    fig, ax = plt.subplots(figsize=figsize)
    ax.boxplot(
        array,
        notch=True,
        vert=False,
        widths=0.7,
        patch_artist=True,
        boxprops=dict(facecolor="lightblue"),
    )
    readable_xticklabels(ax, max_num_xticks=100)
    if xmin < xmax:
        ax.set_xlim(xmin, xmax)
    ax.set_ylabel(" ")
    return figure2html(fig, close_fig=True)


def create_histogram_figure(
    series: Series,
    column: str,
    stats: dict,
    nbins: int | None = None,
    yscale: str = "linear",
    xmin: float | str | None = None,
    xmax: float | str | None = None,
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Creates the HTML code of a histogram figure.

    Args:
        series: Specifies the series/column to analyze.
        column: Specifies the column name.
        stats: Specifies a dictionary with the statistics.
        nbins: Specifies the number of bins in the histogram.
        yscale: Specifies the y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'`` scale is chosen based on the
            distribution.
        xmin: Specifies the minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: Specifies the maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: Specifies the figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        str: The HTML code of the figure.
    """
    array = series.to_numpy()
    if array.size == 0:
        return "<span>&#9888;</span> No figure is generated because the column is empty"
    xmin, xmax = find_range(array, xmin=xmin, xmax=xmax)
    fig, ax = plt.subplots(figsize=figsize)
    ax.hist(array, bins=nbins, range=[xmin, xmax], color="tab:blue", alpha=0.9)
    readable_xticklabels(ax, max_num_xticks=100)
    if xmin < xmax:
        ax.set_xlim(xmin, xmax)
    ax.set_title(f"Distribution of values for column {column}")
    ax.set_ylabel("Number of occurrences")
    if yscale == "auto":
        yscale = auto_yscale_continuous(array=array, nbins=nbins)
    ax.set_yscale(yscale)
    if xmin < stats["q05"] < xmax:
        add_axvline_quantile(ax, x=stats["q05"], label="q0.05 ", horizontalalignment="right")
        ax.axvline(stats["q05"], color="black", linestyle="dashed")
    if xmin < stats["q95"] < xmax:
        add_axvline_quantile(ax, x=stats["q95"], label=" q0.95", horizontalalignment="left")
    # if xmin < stats["median"] < xmax:
    #     add_axvline_median(ax, x=stats["median"])
    add_cdf_plot(ax, array=array, nbins=nbins, xmin=xmin, xmax=xmax)
    ax.legend(
        [Line2D([0], [0], linestyle="none", mfc="black", mec="none", marker="")] * 3,
        [
            f'std={stats["std"]:.2f}',
            f'skewness={stats["skewness"]:.2f}',
            f'kurtosis={stats["kurtosis"]:.2f}',
        ],
    )
    return figure2html(fig, close_fig=True)


def add_axvline_quantile(
    ax: Axes,
    x: float,
    label: str,
    color: str = "black",
    linestyle: str = "dashed",
    horizontalalignment: str = "center",
) -> None:
    ax.axvline(x, color=color, linestyle=linestyle)
    ax.text(
        x,
        0.99,
        label,
        transform=ax.get_xaxis_transform(),
        color=color,
        horizontalalignment=horizontalalignment,
        verticalalignment="top",
    )


def add_axvline_median(
    ax: Axes,
    x: float,
    label: str = "median",
    color: str = "black",
    linestyle: str = "dashed",
    horizontalalignment: str = "center",
) -> None:
    ax.axvline(x, color=color, linestyle=linestyle)
    ax.text(
        x,
        0.99,
        label,
        transform=ax.get_xaxis_transform(),
        color=color,
        horizontalalignment=horizontalalignment,
        verticalalignment="top",
    )


def add_cdf_plot(
    ax: Axes,
    array: np.ndarray,
    nbins: int,
    xmin: float = float("-inf"),
    xmax: float = float("inf"),
    color: str = "tab:red",
) -> None:
    r"""Adds the cumulative distribution function (CDF) plot.

    Args:
        ax:
        array: Specifies the array with the data.
        nbins: Specifies the number of bins to use to plot the CDF.
        xmin: Specifies the minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: Specifies the maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        color: Specifies the plot color.
    """
    nbins = nbins or 1000
    array = nonnan(array)
    nleft = array[array < xmin].size
    nright = array[array > xmax].size
    counts, edges = np.histogram(array[np.logical_and(array >= xmin, array <= xmax)], bins=nbins)
    cdf = (np.cumsum(counts) + nleft) / (np.sum(counts) + nleft + nright)
    x = [(left + right) * 0.5 for left, right in zip(edges[:-1], edges[1:])]
    ax = ax.twinx()
    ax.tick_params(axis="y", labelcolor=color)
    ax.plot(x, cdf, color=color, label="CDF")
    ax.set_ylim(max(0.0, cdf[0]), min(1.0, cdf[-1]))
    ax.set_ylabel("cumulative distribution function (CDF)", color=color)


def create_stats_table(stats: dict, column: str) -> str:
    r"""Creates the HTML code of the table with statistics.

    Args:
        stats: Specifies a dictionary with the statistics.
        column: Specifies the column name.

    Returns:
        The HTML code of the table.
    """
    return Template(
        """
<details>
    <summary>Statistics</summary>

    <p>The following table shows some statistics about the distribution for column {{column}}.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr><th>stat</th><th>value</th></tr>
        </thead>
        <tbody class="tbody table-group-divider">
            <tr><th>count</th><td {{num_style}}>{{count}}</td></tr>
            <tr><th>mean</th><td {{num_style}}>{{mean}}</td></tr>
            <tr><th>std</th><td {{num_style}}>{{std}}</td></tr>
            <tr><th>skewness</th><td {{num_style}}>{{skewness}}</td></tr>
            <tr><th>kurtosis</th><td {{num_style}}>{{kurtosis}}</td></tr>
            <tr><th>min</th><td {{num_style}}>{{min}}</td></tr>
            <tr><th>quantile 0.1%</th><td {{num_style}}>{{q01}}</td></tr>
            <tr><th>quantile 1%</th><td {{num_style}}>{{q01}}</td></tr>
            <tr><th>quantile 5%</th><td {{num_style}}>{{q05}}</td></tr>
            <tr><th>quantile 10%</th><td {{num_style}}>{{q10}}</td></tr>
            <tr><th>quantile 25%</th><td {{num_style}}>{{q25}}</td></tr>
            <tr><th>median</th><td {{num_style}}>{{median}}</td></tr>
            <tr><th>quantile 75%</th><td {{num_style}}>{{q75}}</td></tr>
            <tr><th>quantile 90%</th><td {{num_style}}>{{q90}}</td></tr>
            <tr><th>quantile 95%</th><td {{num_style}}>{{q95}}</td></tr>
            <tr><th>quantile 99%</th><td {{num_style}}>{{q99}}</td></tr>
            <tr><th>quantile 99.9%</th><td {{num_style}}>{{q99}}</td></tr>
            <tr><th>max</th><td {{num_style}}>{{max}}</td></tr>
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render(
        {
            "column": column,
            "num_style": 'style="text-align: right;"',
            "count": f"{stats['count']:,}",
            "mean": f"{stats['mean']:,.4f}",
            "std": f"{stats['std']:,.4f}",
            "skewness": f"{stats['skewness']:,.4f}",
            "kurtosis": f"{stats['kurtosis']:,.4f}",
            "min": f"{stats['min']:,.4f}",
            "q001": f"{stats['q001']:,.4f}",
            "q01": f"{stats['q01']:,.4f}",
            "q05": f"{stats['q05']:,.4f}",
            "q10": f"{stats['q10']:,.4f}",
            "q25": f"{stats['q25']:,.4f}",
            "median": f"{stats['median']:,.4f}",
            "q75": f"{stats['q75']:,.4f}",
            "q90": f"{stats['q90']:,.4f}",
            "q95": f"{stats['q95']:,.4f}",
            "q99": f"{stats['q99']:,.4f}",
            "q999": f"{stats['q999']:,.4f}",
            "max": f"{stats['max']:,.4f}",
        }
    )



from __future__ import annotations

__all__ = ["ColumnContinuousAdvancedAnalyzer"]

import logging

from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import ColumnContinuousAdvancedSection, EmptySection

logger = logging.getLogger(__name__)


class ColumnContinuousAdvancedAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to show the temporal distribution of
    continuous values.

    Args:
        column: Specifies the column name.
        nbins: Specifies the number of bins in the histogram.
        yscale: Specifies the y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        figsize: Specifies the figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import ColumnContinuousAdvancedAnalyzer
        >>> analyzer = ColumnContinuousAdvancedAnalyzer(column="float")
        >>> analyzer
        ColumnContinuousAdvancedAnalyzer(column=float, nbins=None, yscale=auto, figsize=None)
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(
        self,
        column: str,
        nbins: int | None = None,
        yscale: str = "auto",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._column = column
        self._nbins = nbins
        self._yscale = yscale
        self._figsize = figsize

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, nbins={self._nbins}, "
            f"yscale={self._yscale}, figsize={self._figsize})"
        )

    def analyze(self, df: DataFrame) -> ColumnContinuousAdvancedSection | EmptySection:
        logger.info(f"Analyzing the continuous distribution of {self._column}")
        if self._column not in df:
            logger.info(
                "Skipping temporal continuous distribution analysis because the column "
                f"({self._column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        return ColumnContinuousAdvancedSection(
            column=self._column,
            series=df[self._column],
            nbins=self._nbins,
            yscale=self._yscale,
            figsize=self._figsize,
        )



from __future__ import annotations

__all__ = ["human_byte"]


def human_byte(size: float | int, decimal: int = 2) -> str:
    r"""Computes a human readable string representation of byte sizes.

    Args:
        size (float or int): Specifies the number of bytes.
        decimal (int, optional): Specifies the number of decimal
            digits. Default: ``2``

    Returns:
        str: The human readable string representation of byte sizes.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.utils.format import human_byte
        >>> human_byte(2)
        '2.00 B'
        >>> human_byte(2048)
        '2.00 KB'
        >>> human_byte(2097152)
        '2.00 MB'
    """
    for unit in ["B", "KB", "MB", "GB", "TB", "PB"]:
        if size < 1024.0:
            break
        if unit != "PB":
            size /= 1024.0
    return f"{size:,.{decimal}f} {unit}"



from __future__ import annotations

__all__ = ["FilteredAnalyzer"]

import logging

from coola.utils import str_indent, str_mapping
from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import BaseSection
from flamme.utils import setup_object

logger = logging.getLogger(__name__)


class FilteredAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to find all the value types in each
    column.

    Args:
        query (``str``): Soecifies the query.
        analyzer (``BaseAnalyzer`` or dict): Specifies the analyzer
            or its configuration.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import FilteredAnalyzer, NullValueAnalyzer
        >>> analyzer = FilteredAnalyzer(query="float >= 2.0", analyzer=NullValueAnalyzer())
        >>> analyzer
        FilteredAnalyzer(
          (query): float >= 2.0
          (analyzer): NullValueAnalyzer(figsize=None)
        )
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(self, query: str, analyzer: BaseAnalyzer | dict) -> None:
        self._query = query
        self._analyzer = setup_object(analyzer)

    def __repr__(self) -> str:
        args = str_indent(str_mapping({"query": self._query, "analyzer": self._analyzer}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def analyze(self, df: DataFrame) -> BaseSection:
        logger.info("Filtering the DataFrame...")
        df = df.query(self._query)
        return self._analyzer.analyze(df)



from __future__ import annotations

__all__ = ["ToDatetimeSeriesTransformer"]


import pandas as pd
from pandas import Series

from flamme.transformer.series.base import BaseSeriesTransformer


class ToDatetimeSeriesTransformer(BaseSeriesTransformer):
    r"""Implements a transformer to convert a ``pandas.Series`` to
    datetime type.

    Args:
        **kwargs: Specifies the keyword arguments for
            ``pandas.to_datetime``.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.series import ToDatetime
        >>> transformer = ToDatetime()
        >>> transformer
        ToDatetimeSeriesTransformer()
        >>> series = pd.Series(["2020-1-1", "2020-1-2", "2020-1-31", "2020-12-31", "2021-12-31"])
        >>> series.dtype
        dtype('O')
        >>> series = transformer.transform(series)
        >>> series.dtype
        dtype('<M8[ns]')
    """

    def __init__(self, **kwargs) -> None:
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = ", ".join([f"{key}={value}" for key, value in self._kwargs.items()])
        return f"{self.__class__.__qualname__}({args})"

    def transform(self, series: Series) -> Series:
        return pd.to_datetime(series, **self._kwargs)



from __future__ import annotations

__all__ = ["ColumnTemporalNullValueSection"]

import logging
from collections.abc import Sequence

from jinja2 import Template
from matplotlib import pyplot as plt
from pandas import DataFrame

from flamme.section.base import BaseSection
from flamme.section.null import plot_temporal_null_total, prepare_data
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels

logger = logging.getLogger(__name__)


class ColumnTemporalNullValueSection(BaseSection):
    r"""Implements a section to analyze the temporal distribution of null
    values for a given column.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        figsize (``tuple`` , optional): Specifies the figure size in
            inches. The first dimension is the width and the second is
            the height. Default: ``None``
    """

    def __init__(
        self,
        df: DataFrame,
        column: str,
        dt_column: str,
        period: str,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        if column not in df:
            raise ValueError(
                f"Column {column} is not in the DataFrame (columns:{sorted(df.columns)})"
            )
        if dt_column not in df:
            raise ValueError(
                f"Datetime column {dt_column} is not in the DataFrame (columns:{sorted(df.columns)})"
            )

        self._df = df
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._figsize = figsize

    @property
    def df(self) -> DataFrame:
        r"""``pandas.DataFrame``: The DataFrame to analyze."""
        return self._df

    @property
    def column(self) -> str:
        r"""str: The column to analyze."""
        return self._column

    @property
    def dt_column(self) -> str:
        r"""str: The datetime column."""
        return self._dt_column

    @property
    def period(self) -> str:
        r"""str: The temporal period used to analyze the data."""
        return self._period

    @property
    def figsize(self) -> tuple[float, float]:
        r"""tuple: The figure size in inches. The first dimension is
        the width and the second is the height."""
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            f"Rendering the temporal distribution of null values for column {self._column} "
            f"| datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "dt_column": self._dt_column,
                "figure": create_temporal_null_figure(
                    df=self._df,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                    figsize=self._figsize,
                ),
                "table": create_temporal_null_table(
                    df=self._df,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of null values in column <em>{{column}}</em>.
The column <em>{{dt_column}}</em> is used as the temporal column.

{{figure}}

{{table}}
<p style="margin-top: 1rem;">
"""


def create_temporal_null_figure(
    df: DataFrame,
    column: str,
    dt_column: str,
    period: str,
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Creates a HTML representation of a figure with the temporal null
    value distribution.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        figsize (``tuple`` , optional): Specifies the figure size in
            inches. The first dimension is the width and the second is
            the height. Default: ``None``

    Returns:
        str: The HTML representation of the figure.
    """
    if df.shape[0] == 0:
        return ""

    num_nulls, total, labels = prepare_data(
        df=df, column=column, dt_column=dt_column, period=period
    )

    fig, ax = plt.subplots(figsize=figsize)
    plot_temporal_null_total(ax=ax, labels=labels, num_nulls=num_nulls, total=total)
    readable_xticklabels(ax, max_num_xticks=100)
    return figure2html(fig, close_fig=True)


def create_temporal_null_table(df: DataFrame, column: str, dt_column: str, period: str) -> str:
    r"""Creates a HTML representation of a table with the temporal
    distribution of null values.

    Args:
        df (``DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.

    Returns:
        str: The HTML representation of the table.
    """
    if df.shape[0] == 0:
        return ""
    num_nulls, totals, labels = prepare_data(
        df=df, column=column, dt_column=dt_column, period=period
    )
    rows = []
    for label, num_null, total in zip(labels, num_nulls, totals):
        rows.append(create_temporal_null_table_row(label=label, num_nulls=num_null, total=total))
    return Template(
        """
<details>
    <summary>Statistics per period</summary>

    <p>The following table shows some statistics for each period of column {{column}}.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr>
                <th>period</th>
                <th>number of null values</th>
                <th>number of non-null values</th>
                <th>total number of values</th>
                <th>percentage of null values</th>
                <th>percentage of non-null values</th>
            </tr>
        </thead>
        <tbody class="tbody table-group-divider">
            {{rows}}
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render({"rows": "\n".join(rows), "column": column, "period": period})


def create_temporal_null_table_row(label: str, num_nulls: int, total: int) -> str:
    r"""Creates the HTML code of a new table row.

    Args:
        row ("pd.core.frame.Pandas"): Specifies a DataFrame row.

    Returns:
        str: The HTML code of a row.
    """
    num_non_nulls = total - num_nulls
    return Template(
        """<tr>
    <th>{{label}}</th>
    <td {{num_style}}>{{num_nulls}}</td>
    <td {{num_style}}>{{num_non_nulls}}</td>
    <td {{num_style}}>{{total}}</td>
    <td {{num_style}}>{{num_nulls_pct}}</td>
    <td {{num_style}}>{{num_non_nulls_pct}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "label": label,
            "num_nulls": f"{num_nulls:,}",
            "num_non_nulls": f"{num_non_nulls:,}",
            "total": f"{total:,}",
            "num_nulls_pct": f"{100 * num_nulls / total:.2f}%",
            "num_non_nulls_pct": f"{100 * num_non_nulls / total:.2f}%",
        }
    )



from __future__ import annotations

__all__ = ["DataTypeAnalyzer"]

import logging

from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import DataTypeSection
from flamme.utils.dtype import df_column_types

logger = logging.getLogger(__name__)


class DataTypeAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to find all the value types in each
    column.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import DataTypeAnalyzer
        >>> analyzer = DataTypeAnalyzer()
        >>> analyzer
        DataTypeAnalyzer()
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def analyze(self, df: DataFrame) -> DataTypeSection:
        logger.info("Analyzing the data types...")
        return DataTypeSection(dtypes=df.dtypes.to_dict(), types=df_column_types(df))



from __future__ import annotations

__all__ = ["ColumnSelectionDataFrameTransformer"]

from collections.abc import Sequence

from pandas import DataFrame

from flamme.transformer.df.base import BaseDataFrameTransformer


class ColumnSelectionDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implements a ``pandas.DataFrame`` transformer to select a subset
    of columns.

    Args:
        columns (``Sequence``): Specifies the columns to keep.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.df import ColumnSelection
        >>> transformer = ColumnSelection(columns=["col1", "col2"])
        >>> transformer
        ColumnSelectionDataFrameTransformer(columns=['col1', 'col2'])
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": ["2020-1-1", "2020-1-2", "2020-1-31", "2020-12-31", None],
        ...         "col2": [1, 2, 3, 4, 5],
        ...         "col3": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> df = transformer.transform(df)
        >>> df
                 col1  col2
        0    2020-1-1     1
        1    2020-1-2     2
        2   2020-1-31     3
        3  2020-12-31     4
        4        None     5
    """

    def __init__(self, columns: Sequence[str]) -> None:
        self._columns = list(columns)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(columns={self._columns})"

    def transform(self, df: DataFrame) -> DataFrame:
        for col in self._columns:
            if col not in df:
                raise RuntimeError(
                    f"Column {col} is not in the DataFrame (columns:{sorted(df.columns)})"
                )
        return df[self._columns].copy()



from __future__ import annotations

__all__ = ["ColumnTemporalContinuousSection"]

import logging
from collections.abc import Sequence

import numpy as np
import pandas as pd
from jinja2 import Template
from matplotlib import pyplot as plt
from pandas import DataFrame

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    auto_yscale_continuous,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html, readable_xticklabels
from flamme.utils.mathnan import remove_nan

logger = logging.getLogger(__name__)


class ColumnTemporalContinuousSection(BaseSection):
    r"""Implements a section that analyzes the temporal distribution of a
    column with continuous values.

    Args:
        df: Specifies the DataFrame to analyze.
        column: Specifies the column to analyze.
        dt_column: Specifies the datetime column used to analyze
            the temporal distribution.
        period: Specifies the temporal period e.g. monthly or daily.
        yscale: Specifies the y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        figsize: Specifies the figure size in inches. The first
            dimension is the width and the second is the height.
    """

    def __init__(
        self,
        df: DataFrame,
        column: str,
        dt_column: str,
        period: str,
        yscale: str = "auto",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._df = df
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._yscale = yscale
        self._figsize = figsize

    @property
    def column(self) -> str:
        return self._column

    @property
    def dt_column(self) -> str:
        return self._dt_column

    @property
    def yscale(self) -> str:
        return self._yscale

    @property
    def period(self) -> str:
        return self._period

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""tuple: The individual figure size in pixels. The first
        dimension is the width and the second is the height."""
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            f"Rendering the temporal continuous distribution of {self._column} | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(self._create_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "dt_column": self._dt_column,
                "period": self._period,
                "figure": create_temporal_figure(
                    df=self._df,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                    yscale=self._yscale,
                    figsize=self._figsize,
                ),
                "table": create_temporal_table(
                    df=self._df,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_template(self) -> str:
        return """
<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of column <em>{{column}}</em>
by using the column <em>{{dt_column}}</em>.

{{figure}}

{{table}}
<p style="margin-top: 1rem;">
"""


def create_temporal_figure(
    df: DataFrame,
    column: str,
    dt_column: str,
    period: str,
    yscale: str = "auto",
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Creates a HTML representation of a figure with the temporal value
    distribution.

    Args:
        df: Specifies the DataFrame to analyze.
        column: Specifies the column to analyze.
        dt_column: Specifies the datetime column used to analyze
            the temporal distribution.
        period: Specifies the temporal period e.g. monthly or daily.
        yscale: Specifies the y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        figsize: Specifies the figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The HTML representation of the figure.
    """
    if df.shape[0] == 0:
        return "<span>&#9888;</span> No figure is generated because the column is empty"
    array = df[column].to_numpy(dtype=float)
    df = df[[column, dt_column]].copy()
    df[dt_column] = df[dt_column].dt.to_period(period).astype(str)
    df_group = (
        df.groupby(dt_column)[column]
        .apply(list)
        .reset_index(name=column)
        .sort_values(by=[dt_column])
    )

    data = [remove_nan(x) for x in df_group[column].tolist()]
    labels = df_group[dt_column].tolist()
    fig, ax = plt.subplots(figsize=figsize)
    ax.boxplot(
        data,
        notch=True,
        vert=True,
        widths=0.7,
        patch_artist=True,
        boxprops=dict(facecolor="lightblue"),
    )
    ax.set_ylim(np.nanmin(array), np.nanmax(array))
    ax.set_xticks(np.arange(len(labels)), labels=labels)
    ax.set_title(f"Distribution of values for column {column}")
    if yscale == "auto":
        yscale = auto_yscale_continuous(array=array, nbins=100)
    ax.set_yscale(yscale)
    readable_xticklabels(ax)
    return figure2html(fig, close_fig=True)


def create_temporal_table(df: DataFrame, column: str, dt_column: str, period: str) -> str:
    r"""Creates a HTML representation of a table with some statistics
    about the temporal value distribution.

    Args:
        df (``DataFrame``): Specifies the DataFrame to analyze.
        column (str): Specifies the column to analyze.
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.

    Returns:
        str: The HTML representation of the table.
    """
    if df.shape[0] == 0:
        return "<span>&#9888;</span> No table is generated because the column is empty"
    df = df[[column, dt_column]].copy()
    dt_col = "__datetime__"
    df[dt_col] = df[dt_column].dt.to_period(period)
    df_stats = (
        df.groupby(dt_col)[column]
        .agg(
            [
                "count",
                "mean",
                "median",
                "min",
                "max",
                "std",
                ("q01", lambda x: x.quantile(0.01)),
                ("q05", lambda x: x.quantile(0.05)),
                ("q10", lambda x: x.quantile(0.1)),
                ("q25", lambda x: x.quantile(0.25)),
                ("q75", lambda x: x.quantile(0.75)),
                ("q90", lambda x: x.quantile(0.9)),
                ("q95", lambda x: x.quantile(0.95)),
                ("q99", lambda x: x.quantile(0.99)),
            ]
        )
        .sort_index()
    )

    rows = []
    for row in df_stats.itertuples():
        rows.append(create_temporal_table_row(row))
    return Template(
        """
<details>
    <summary>Statistics per period</summary>

    <p>The following table shows some statistics for each period of column {{column}}.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr>
                <th>column</th>
                <th>count</th>
                <th>mean</th>
                <th>std</th>
                <th>min</th>
                <th>quantile 1%</th>
                <th>quantile 5%</th>
                <th>quantile 10%</th>
                <th>quantile 25%</th>
                <th>median</th>
                <th>quantile 75%</th>
                <th>quantile 90%</th>
                <th>quantile 95%</th>
                <th>quantile 99%</th>
                <th>max</th>
            </tr>
        </thead>
        <tbody class="tbody table-group-divider">
            {{rows}}
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render({"rows": "\n".join(rows), "column": column, "period": period})


def create_temporal_table_row(row: pd.core.frame.Pandas) -> str:
    r"""Creates the HTML code of a new table row.

    Args:
        row: Specifies a DataFrame row.

    Returns:
        The HTML code of a row.
    """
    return Template(
        """<tr>
    <th>{{datetime}}</th>
    <td {{num_style}}>{{count}}</td>
    <td {{num_style}}>{{mean}}</td>
    <td {{num_style}}>{{std}}</td>
    <td {{num_style}}>{{min}}</td>
    <td {{num_style}}>{{q01}}</td>
    <td {{num_style}}>{{q05}}</td>
    <td {{num_style}}>{{q10}}</td>
    <td {{num_style}}>{{q25}}</td>
    <td {{num_style}}>{{median}}</td>
    <td {{num_style}}>{{q75}}</td>
    <td {{num_style}}>{{q90}}</td>
    <td {{num_style}}>{{q95}}</td>
    <td {{num_style}}>{{q99}}</td>
    <td {{num_style}}>{{max}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "datetime": row.Index,
            "count": f"{row.count:,}",
            "mean": f"{row.mean:,.4f}",
            "median": f"{row.median:,.4f}",
            "min": f"{row.min:,.4f}",
            "max": f"{row.max:,.4f}",
            "std": f"{row.std:,.4f}",
            "q01": f"{row.q01:,.4f}",
            "q05": f"{row.q05:,.4f}",
            "q10": f"{row.q10:,.4f}",
            "q25": f"{row.q25:,.4f}",
            "q75": f"{row.q75:,.4f}",
            "q90": f"{row.q90:,.4f}",
            "q95": f"{row.q95:,.4f}",
            "q99": f"{row.q99:,.4f}",
        }
    )



from __future__ import annotations

__all__ = ["BaseReporter", "is_reporter_config", "setup_reporter"]

import logging
from abc import ABC

from objectory import AbstractFactory
from objectory.utils import is_object_config

logger = logging.getLogger(__name__)


class BaseReporter(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to compute a HTML report.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.analyzer import NullValueAnalyzer
        >>> from flamme.ingestor import ParquetIngestor
        >>> from flamme.transformer.df import SequentialDataFrameTransformer
        >>> from flamme.reporter import Reporter
        >>> reporter = Reporter(
        ...     ingestor=ParquetIngestor("/path/to/data.parquet"),
        ...     transformer=SequentialDataFrameTransformer(transformers=[]),
        ...     analyzer=NullValueAnalyzer(),
        ...     report_path="/path/to/report.html",
        ... )
        >>> reporter
        Reporter(
          (ingestor): ParquetIngestor(path=/path/to/data.parquet)
          (transformer): SequentialDataFrameTransformer()
          (analyzer): NullValueAnalyzer(figsize=None)
          (report_path): /path/to/report.html
          (max_toc_depth): 6
        )
        >>> report = reporter.compute()  # doctest: +SKIP
    """

    def compute(self) -> None:
        r"""Computes a HTML report.

        Example usage:

        .. code-block:: pycon

            >>> from flamme.analyzer import NullValueAnalyzer
            >>> from flamme.ingestor import ParquetIngestor
            >>> from flamme.transformer.df import SequentialDataFrameTransformer
            >>> from flamme.reporter import Reporter
            >>> reporter = Reporter(
            ...     ingestor=ParquetIngestor("/path/to/data.parquet"),
            ...     transformer=SequentialDataFrameTransformer(transformers=[]),
            ...     analyzer=NullValueAnalyzer(figsize=None),
            ...     report_path="/path/to/report.html",
            ... )
            >>> report = reporter.compute()  # doctest: +SKIP
        """


def is_reporter_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseReporter``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config (dict): Specifies the configuration to check.

    Returns:
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseReporter`` object.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.reporter import is_reporter_config
        >>> is_reporter_config(
        ...     {
        ...         "_target_": "flamme.reporter.Reporter",
        ...         "ingestor": {
        ...             "_target_": "flamme.ingestor.CsvIngestor",
        ...             "path": "/path/to/data.csv",
        ...         },
        ...         "transformer": {
        ...             "_target_": "flamme.transformer.df.ToNumeric",
        ...             "columns": ["col1", "col3"],
        ...         },
        ...         "analyzer": {"_target_": "flamme.analyzer.NullValueAnalyzer"},
        ...         "report_path": "/path/to/report.html",
        ...     }
        ... )
        True
    """
    return is_object_config(config, BaseReporter)


def setup_reporter(
    reporter: BaseReporter | dict,
) -> BaseReporter:
    r"""Sets up an reporter.

    The reporter is instantiated from its configuration
    by using the ``BaseReporter`` factory function.

    Args:
        reporter (``BaseReporter`` or dict): Specifies an
            reporter or its configuration.

    Returns:
        ``BaseReporter``: An instantiated reporter.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.reporter import setup_reporter
        >>> reporter = setup_reporter(
        ...     {
        ...         "_target_": "flamme.reporter.Reporter",
        ...         "ingestor": {
        ...             "_target_": "flamme.ingestor.CsvIngestor",
        ...             "path": "/path/to/data.csv",
        ...         },
        ...         "transformer": {
        ...             "_target_": "flamme.transformer.df.ToNumeric",
        ...             "columns": ["col1", "col3"],
        ...         },
        ...         "analyzer": {"_target_": "flamme.analyzer.NullValueAnalyzer"},
        ...         "report_path": "/path/to/report.html",
        ...     }
        ... )
        >>> reporter
        Reporter(
          (ingestor): CsvIngestor(path=/path/to/data.csv)
          (transformer): ToNumericDataFrameTransformer(columns=('col1', 'col3'))
          (analyzer): NullValueAnalyzer(figsize=None)
          (report_path): /path/to/report.html
          (max_toc_depth): 6
        )
    """
    if isinstance(reporter, dict):
        logger.info("Initializing an reporter from its configuration... ")
        reporter = BaseReporter.factory(**reporter)
    if not isinstance(reporter, BaseReporter):
        logger.warning(f"reporter is not a `BaseReporter` (received: {type(reporter)})")
    return reporter



from __future__ import annotations

__all__ = ["TemporalNullValueAnalyzer", "NullValueAnalyzer"]

import logging

import numpy as np
from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import EmptySection
from flamme.section.null import NullValueSection, TemporalNullValueSection

logger = logging.getLogger(__name__)


class NullValueAnalyzer(BaseAnalyzer):
    r"""Implements a null value analyzer.

    Args:
        figsize (``tuple`` , optional): Specifies the figure size in
            inches. The first dimension is the width and the second is
            the height. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import NullValueAnalyzer
        >>> analyzer = NullValueAnalyzer()
        >>> analyzer
        NullValueAnalyzer(figsize=None)
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(self, figsize: tuple[float, float] | None = None) -> None:
        self._figsize = figsize

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(figsize={self._figsize})"

    def analyze(self, df: DataFrame) -> NullValueSection:
        logger.info("Analyzing the null value distribution of all columns...")
        return NullValueSection(
            columns=list(df.columns),
            null_count=df.isnull().sum().to_frame("count")["count"].to_numpy(),
            total_count=np.full((df.shape[1],), df.shape[0]),
            figsize=self._figsize,
        )


class TemporalNullValueAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to show the temporal distribution of null
    values.

    Args:
        dt_column (str): Specifies the datetime column used to analyze
            the temporal distribution.
        period (str): Specifies the temporal period e.g. monthly or
            daily.
        ncols (int, optional): Specifies the number of columns.
            Default: ``2``
        figsize (``tuple``, optional): Specifies the figure size in
            inches. The first dimension is the width and the second is
            the height. Default: ``(7, 5)``

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import TemporalNullValueAnalyzer
        >>> analyzer = TemporalNullValueAnalyzer("datetime", period="M")
        >>> analyzer
        TemporalNullValueAnalyzer(dt_column=datetime, period=M, ncols=2, figsize=(7, 5))
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...         "datetime": pd.to_datetime(
        ...             ["2020-01-03", "2020-02-03", "2020-03-03", "2020-04-03"]
        ...         ),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(
        self,
        dt_column: str,
        period: str,
        ncols: int = 2,
        figsize: tuple[float, float] = (7, 5),
    ) -> None:
        self._dt_column = dt_column
        self._period = period
        self._ncols = ncols
        self._figsize = figsize

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(dt_column={self._dt_column}, period={self._period}, "
            f"ncols={self._ncols}, figsize={self._figsize})"
        )

    def analyze(self, df: DataFrame) -> TemporalNullValueSection | EmptySection:
        logger.info(
            "Analyzing the temporal null value distribution of all columns | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        if self._dt_column not in df:
            logger.info(
                "Skipping monthly null value analysis because the datetime column "
                f"({self._dt_column}) is not in the DataFrame: {sorted(df.columns)}"
            )
            return EmptySection()
        return TemporalNullValueSection(
            df=df,
            dt_column=self._dt_column,
            period=self._period,
            ncols=self._ncols,
            figsize=self._figsize,
        )



from __future__ import annotations

__all__ = [
    "BaseAnalyzer",
    "ChoiceAnalyzer",
    "ColumnContinuousAdvancedAnalyzer",
    "ColumnContinuousAnalyzer",
    "ColumnDiscreteAnalyzer",
    "ColumnSubsetAnalyzer",
    "ColumnTemporalContinuousAnalyzer",
    "ColumnTemporalDiscreteAnalyzer",
    "ColumnTemporalNullValueAnalyzer",
    "DataTypeAnalyzer",
    "DuplicatedRowAnalyzer",
    "FilteredAnalyzer",
    "GlobalTemporalNullValueAnalyzer",
    "MappingAnalyzer",
    "MarkdownAnalyzer",
    "MostFrequentValuesAnalyzer",
    "NullValueAnalyzer",
    "TemporalNullValueAnalyzer",
    "is_analyzer_config",
    "setup_analyzer",
]

from flamme.analyzer.base import BaseAnalyzer, is_analyzer_config, setup_analyzer
from flamme.analyzer.choice import ChoiceAnalyzer
from flamme.analyzer.column import ColumnSubsetAnalyzer
from flamme.analyzer.continuous import (
    ColumnContinuousAnalyzer,
    ColumnTemporalContinuousAnalyzer,
)
from flamme.analyzer.continuous_advanced import ColumnContinuousAdvancedAnalyzer
from flamme.analyzer.discrete import (
    ColumnDiscreteAnalyzer,
    ColumnTemporalDiscreteAnalyzer,
)
from flamme.analyzer.dtype import DataTypeAnalyzer
from flamme.analyzer.duplicate import DuplicatedRowAnalyzer
from flamme.analyzer.filter import FilteredAnalyzer
from flamme.analyzer.mapping import MappingAnalyzer
from flamme.analyzer.markdown import MarkdownAnalyzer
from flamme.analyzer.most_frequent import MostFrequentValuesAnalyzer
from flamme.analyzer.null import NullValueAnalyzer, TemporalNullValueAnalyzer
from flamme.analyzer.null_temp_col import ColumnTemporalNullValueAnalyzer
from flamme.analyzer.null_temp_global import GlobalTemporalNullValueAnalyzer



from __future__ import annotations

__all__ = ["BaseIngestor", "is_ingestor_config", "setup_ingestor"]

import logging
from abc import ABC

from objectory import AbstractFactory
from objectory.utils import is_object_config
from pandas import DataFrame

logger = logging.getLogger(__name__)


class BaseIngestor(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to implement a DataFrame ingestor.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.ingestor import ParquetIngestor
        >>> ingestor = ParquetIngestor(path="/path/to/df.parquet")
        >>> ingestor
        ParquetIngestor(path=/path/to/df.parquet)
        >>> df = ingestor.ingest()  # doctest: +SKIP
    """

    def ingest(self) -> DataFrame:
        r"""Ingests a DataFrame.

        Returns:
            ``pandas.DataFrame``: The ingested DataFrame.

        Example usage:

        .. code-block:: pycon

            >>> from flamme.ingestor import ParquetIngestor
            >>> ingestor = ParquetIngestor(path="/path/to/df.parquet")
            >>> df = ingestor.ingest()  # doctest: +SKIP
        """


def is_ingestor_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseIngestor``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config (dict): Specifies the configuration to check.

    Returns:
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseIngestor`` object.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.ingestor import is_ingestor_config
        >>> is_ingestor_config(
        ...     {"_target_": "flamme.ingestor.CsvIngestor", "path": "/path/to/data.csv"}
        ... )
        True
    """
    return is_object_config(config, BaseIngestor)


def setup_ingestor(
    ingestor: BaseIngestor | dict,
) -> BaseIngestor:
    r"""Sets up an ingestor.

    The ingestor is instantiated from its configuration
    by using the ``BaseIngestor`` factory function.

    Args:
        ingestor (``BaseIngestor`` or dict): Specifies an
            ingestor or its configuration.

    Returns:
        ``BaseIngestor``: An instantiated ingestor.

    Example usage:

    .. code-block:: pycon

        >>> from flamme.ingestor import setup_ingestor
        >>> ingestor = setup_ingestor(
        ...     {"_target_": "flamme.ingestor.CsvIngestor", "path": "/path/to/data.csv"}
        ... )
        >>> ingestor
        CsvIngestor(path=/path/to/data.csv)
    """
    if isinstance(ingestor, dict):
        logger.info("Initializing an ingestor from its configuration... ")
        ingestor = BaseIngestor.factory(**ingestor)
    if not isinstance(ingestor, BaseIngestor):
        logger.warning(f"ingestor is not a `BaseIngestor` (received: {type(ingestor)})")
    return ingestor



from __future__ import annotations

__all__ = ["SequentialDataFrameTransformer"]

from collections.abc import Sequence

from coola.utils import str_indent, str_sequence
from pandas import DataFrame

from flamme.transformer.df.base import (
    BaseDataFrameTransformer,
    setup_dataframe_transformer,
)


class SequentialDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implements a ``pandas.DataFrame`` transformer to apply
    sequentially several transformers.

    Args:
        transformers (``Sequence``): Specifies the transformers or
            their configurations.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.df import (
        ...     Sequential,
        ...     StripString,
        ...     ToNumeric,
        ... )
        >>> transformer = Sequential(
        ...     [
        ...         StripString(columns=["col1", "col3"]),
        ...         ToNumeric(columns=["col1", "col2"]),
        ...     ]
        ... )
        >>> transformer
        SequentialDataFrameTransformer(
          (0): StripStringDataFrameTransformer(columns=('col1', 'col3'))
          (1): ToNumericDataFrameTransformer(columns=('col1', 'col2'))
        )
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["a ", " b", "  c  ", "d", "e"],
        ...         "col4": ["a ", " b", "  c  ", "d", "e"],
        ...     }
        ... )
        >>> df
           col1 col2   col3   col4
        0     1    1     a      a
        1     2    2      b      b
        2     3    3    c      c
        3     4    4      d      d
        4     5    5      e      e
        >>> df = transformer.transform(df)
        >>> df
           col1 col2 col3   col4
        0     1    1    a     a
        1     2    2    b      b
        2     3    3    c    c
        3     4    4    d      d
        4     5    5    e      e
    """

    def __init__(self, transformers: Sequence[BaseDataFrameTransformer | dict]) -> None:
        self._transformers = tuple(
            setup_dataframe_transformer(transformer) for transformer in transformers
        )

    def __repr__(self) -> str:
        args = ""
        if self._transformers:
            args = f"\n  {str_indent(str_sequence(self._transformers))}\n"
        return f"{self.__class__.__qualname__}({args})"

    def transform(self, df: DataFrame) -> DataFrame:
        for transformer in self._transformers:
            df = transformer.transform(df)
        return df



from __future__ import annotations

__all__ = ["SequentialSeriesTransformer"]

from collections.abc import Sequence

from coola.utils import str_indent, str_sequence
from pandas import Series

from flamme.transformer.series import BaseSeriesTransformer, setup_series_transformer


class SequentialSeriesTransformer(BaseSeriesTransformer):
    r"""Implements a ``pandas.Series`` transformer to apply sequentially
    several transformers.

    Args:
        transformers (``Sequence``): Specifies the transformers or
            their configurations.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.series import Sequential, StripString, ToNumeric
        >>> transformer = Sequential([StripString(), ToNumeric()])
        >>> transformer
        SequentialSeriesTransformer(
          (0): StripStringSeriesTransformer()
          (1): ToNumericSeriesTransformer()
        )
        >>> series = pd.Series([" 1", "2 ", " 3 ", "4", "5"])
        >>> series = transformer.transform(series)
        >>> series
        0    1
        1    2
        2    3
        3    4
        4    5
        dtype: int64
    """

    def __init__(self, transformers: Sequence[BaseSeriesTransformer | dict]) -> None:
        self._transformers = tuple(
            setup_series_transformer(transformer) for transformer in transformers
        )

    def __repr__(self) -> str:
        args = ""
        if self._transformers:
            args = f"\n  {str_indent(str_sequence(self._transformers))}\n"
        return f"{self.__class__.__qualname__}({args})"

    def transform(self, series: Series) -> Series:
        for transformer in self._transformers:
            series = transformer.transform(series)
        return series



from __future__ import annotations

__all__ = ["df_column_types", "series_column_types"]

from pandas import DataFrame, Series


def df_column_types(df: DataFrame) -> dict[str, set]:
    r"""Computes the value types per column.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.

    Returns:
        dict: A dictionary with the value types for each column.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.utils.dtype import df_column_types
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...     }
        ... )
        >>> coltypes = df_column_types(df)
        >>> coltypes
        {'int': {<class 'float'>}, 'float': {<class 'float'>}}
    """
    types = {}
    for col in df:
        types[col] = series_column_types(df[col])
    return types


def series_column_types(series: Series) -> set[type]:
    r"""Computes the value types in a ``pandas.Series``.

    Args:
        df (``pandas.DataFrame``): Specifies the DataFrame to analyze.

    Returns:
        dict: A dictionary with the value types for each column.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.utils.dtype import series_column_types
        >>> coltypes = series_column_types(pd.Series([1.2, 4.2, np.nan, 2.2]))
        >>> coltypes
        {<class 'float'>}
    """
    return {type(x) for x in series.tolist()}



from __future__ import annotations

__all__ = ["MappingAnalyzer"]

from collections.abc import Mapping

from coola.utils import str_indent, str_mapping
from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer, setup_analyzer
from flamme.section import SectionDict


class MappingAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to analyze multiple analyzers.

    Args:
        analyzers (``Mapping``): Specifies the mappings to analyze.
            The key of each analyzer is used to organize the metrics
            and report.

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import (
        ...     FilteredAnalyzer,
        ...     NullValueAnalyzer,
        ...     DuplicatedRowAnalyzer,
        ...     MappingAnalyzer,
        ... )
        >>> analyzer = MappingAnalyzer(
        ...     {"null": NullValueAnalyzer(), "duplicate": DuplicatedRowAnalyzer()}
        ... )
        >>> analyzer
        MappingAnalyzer(
          (null): NullValueAnalyzer(figsize=None)
          (duplicate): DuplicatedRowAnalyzer(columns=None, figsize=None)
        )
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(self, analyzers: Mapping[str, BaseAnalyzer | dict]) -> None:
        self._analyzers = {name: setup_analyzer(analyzer) for name, analyzer in analyzers.items()}

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_mapping(self._analyzers))}\n)"

    @property
    def analyzers(self) -> dict[str, BaseAnalyzer]:
        return self._analyzers

    def analyze(self, df: DataFrame) -> SectionDict:
        return SectionDict(
            {name: analyzer.analyze(df) for name, analyzer in self._analyzers.items()}
        )

    def add_analyzer(self, key: str, analyzer: BaseAnalyzer, replace_ok: bool = False) -> None:
        r"""Adds an analyzer to the current analyzer.

        Args:
            key: Specifies the key of the analyzer.
            analyzer: Specifies the analyzer to add.
            replace_ok: If ``False``, ``KeyError`` is raised if an
                analyzer with the same key exists. If ``True``,
                the new analyzer will replace the existing analyzer.

        Raises:
            KeyError: if an  analyzer with the same key exists.

        Example usage:

        ```pycon
        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import MappingAnalyzer, NullValueAnalyzer, DuplicatedRowAnalyzer
        >>> analyzer = MappingAnalyzer({"null": NullValueAnalyzer()})
        >>> analyzer.add_analyzer("duplicate", DuplicatedRowAnalyzer())
        >>> analyzer
        MappingAnalyzer(
          (null): NullValueAnalyzer(figsize=None)
          (duplicate): DuplicatedRowAnalyzer(columns=None, figsize=None)
        )
        >>> df = pd.DataFrame(
        ...     {
        ...         "int": np.array([np.nan, 1, 0, 1]),
        ...         "float": np.array([1.2, 4.2, np.nan, 2.2]),
        ...         "str": np.array(["A", "B", None, np.nan]),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)

        ```
        """
        if key in self._analyzers and not replace_ok:
            raise KeyError(
                f"`{key}` is already used to register an analyzer. "
                "Use `replace_ok=True` to replace the analyzer"
            )
        self._analyzers[key] = analyzer



from __future__ import annotations

__all__ = ["Reporter"]

import logging
from pathlib import Path

from coola.utils import str_indent, str_mapping

from flamme.analyzer.base import BaseAnalyzer, setup_analyzer
from flamme.ingestor.base import BaseIngestor, setup_ingestor
from flamme.reporter.base import BaseReporter
from flamme.reporter.utils import create_html_report
from flamme.transformer.df.base import (
    BaseDataFrameTransformer,
    setup_dataframe_transformer,
)
from flamme.utils.io import save_text
from flamme.utils.path import sanitize_path

logger = logging.getLogger(__name__)


class Reporter(BaseReporter):
    r"""Implements a simple reporter.

    Args:
        ingestor (``BaseIngestor`` or dict): Specifies the ingestor
            or its configuration.
        transformer (``BaseDataFrameTransformer`` or dict): Specifies
            a ``pandas.DataFrame`` transformer or its configuration.
        ingestor (``BaseAnalyzer`` or dict): Specifies the analyzer
            or its configuration.
        report_path (``Path`` or str): Specifies the path where to
            save the HTML report.
        max_toc_depth (int, optional): Specifies the maximum level
            to show in the table of content. Default: ``6``

    Example usage:

    .. code-block:: pycon

        >>> from flamme.analyzer import NullValueAnalyzer
        >>> from flamme.ingestor import ParquetIngestor
        >>> from flamme.transformer.df import SequentialDataFrameTransformer
        >>> from flamme.reporter import Reporter
        >>> reporter = Reporter(
        ...     ingestor=ParquetIngestor("/path/to/data.parquet"),
        ...     transformer=SequentialDataFrameTransformer(transformers=[]),
        ...     analyzer=NullValueAnalyzer(),
        ...     report_path="/path/to/report.html",
        ... )
        >>> report = reporter.compute()  # doctest: +SKIP
    """

    def __init__(
        self,
        ingestor: BaseIngestor | dict,
        transformer: BaseDataFrameTransformer | dict,
        analyzer: BaseAnalyzer | dict,
        report_path: Path | str,
        max_toc_depth: int = 6,
    ) -> None:
        self._ingestor = setup_ingestor(ingestor)
        logger.info(f"ingestor:\n{ingestor}")
        self._transformer = setup_dataframe_transformer(transformer)
        logger.info(f"transformer:\n{transformer}")
        self._analyzer = setup_analyzer(analyzer)
        logger.info(f"analyzer:\n{analyzer}")
        self._report_path = sanitize_path(report_path)
        self._max_toc_depth = int(max_toc_depth)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "ingestor": self._ingestor,
                    "transformer": self._transformer,
                    "analyzer": self._analyzer,
                    "report_path": self._report_path,
                    "max_toc_depth": self._max_toc_depth,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def compute(self) -> None:
        logger.info("Ingesting the DataFrame...")
        df = self._ingestor.ingest()
        logger.info(f"Transforming the DataFrame {df.shape}...")
        df = self._transformer.transform(df)
        logger.info(f"Analyzing the DataFrame {df.shape}...")
        section = self._analyzer.analyze(df)
        logger.info("Creating the HTML report...")
        report = create_html_report(
            toc=section.render_html_toc(max_depth=self._max_toc_depth),
            body=section.render_html_body(),
        )
        logger.info(f"Saving HTML report at {self._report_path}...")
        save_text(report, self._report_path)



from __future__ import annotations

__all__ = ["DuplicatedRowAnalyzer"]

import logging
from collections.abc import Sequence

from pandas import DataFrame

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import DuplicatedRowSection

logger = logging.getLogger(__name__)


class DuplicatedRowAnalyzer(BaseAnalyzer):
    r"""Implements an analyzer to show the number of duplicated rows.

    Args:
        columns (``Sequence`` or ``None``): Specifies the columns used
            to compute the duplicated rows. ``None`` means all the
            columns. Default: ``None``
        figsize (``tuple`` or ``None``, optional): Specifies the figure
            size in inches. The first dimension is the width and the
            second is the height. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> import numpy as np
        >>> import pandas as pd
        >>> from flamme.analyzer import DuplicatedRowAnalyzer
        >>> analyzer = DuplicatedRowAnalyzer()
        >>> analyzer
        DuplicatedRowAnalyzer(columns=None, figsize=None)
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": np.array([0, 1, 0, 1]),
        ...         "col2": np.array([1, 0, 1, 0]),
        ...         "col3": np.array(
        ...             [
        ...                 1,
        ...                 1,
        ...                 1,
        ...                 1,
        ...             ]
        ...         ),
        ...     }
        ... )
        >>> section = analyzer.analyze(df)
    """

    def __init__(
        self,
        columns: Sequence[str] | None = None,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._columns = columns
        self._figsize = figsize

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(columns={self._columns}, figsize={self._figsize})"

    def analyze(self, df: DataFrame) -> DuplicatedRowSection:
        logger.info(f"Analyzing the duplicated rows section using the columns: {self._columns}")
        return DuplicatedRowSection(df=df, columns=self._columns, figsize=self._figsize)



from __future__ import annotations

__all__ = ["ToNumericDataFrameTransformer"]

from collections.abc import Sequence

import pandas as pd
from pandas import DataFrame
from tqdm import tqdm

from flamme.transformer.df.base import BaseDataFrameTransformer


class ToNumericDataFrameTransformer(BaseDataFrameTransformer):
    r"""Implements a transformer to convert some columns to numeric type.

    Args:
        columns (``Sequence``): Specifies the columns to convert.
        **kwargs: Specifies the keyword arguments for
            ``pandas.to_numeric``.

    Example usage:

    .. code-block:: pycon

        >>> import pandas as pd
        >>> from flamme.transformer.df import ToNumeric
        >>> transformer = ToNumeric(columns=["col1", "col3"])
        >>> transformer
        ToNumericDataFrameTransformer(columns=('col1', 'col3'))
        >>> df = pd.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["1", "2", "3", "4", "5"],
        ...         "col4": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> df.dtypes
        col1     int64
        col2    object
        col3    object
        col4    object
        dtype: object
        >>> df = transformer.transform(df)
        >>> df.dtypes
        col1     int64
        col2    object
        col3     int64
        col4    object
        dtype: object
    """

    def __init__(self, columns: Sequence[str], **kwargs) -> None:
        self._columns = tuple(columns)
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = ", ".join([f"{key}={value}" for key, value in self._kwargs.items()])
        if args:
            args = ", " + args
        return f"{self.__class__.__qualname__}(columns={self._columns}{args})"

    def transform(self, df: DataFrame) -> DataFrame:
        for col in tqdm(self._columns, desc="Converting to numeric type"):
            df[col] = pd.to_numeric(df[col], **self._kwargs)
        return df
